@@ -65,7 +65,7 @@ def get_variables() :
     d[i] = c
   return d
 
-def read_calib_tuple(sample, trees, branches, verbose = False, cachedir = None) : 
+def read_calib_tuple(sample, trees, branches, verbose = False, cachedir = None, cut = None) : 
   """ 
   Read calibration sample from the list of files into numpy array. 
 
@@ -95,7 +95,7 @@ def read_calib_tuple(sample, trees, branches, verbose = False, cachedir = None)
     for tree in trees : 
       try : 
         with uproot.open(filename + ":" + tree) as t :
-          arr = t.arrays(branches, library = "pd")[branches].to_numpy()
+          arr = t.arrays(branches, cut, library = "pd")[branches].to_numpy()
 
           if cachedir : 
             outtree = tree.replace("/", "_") # Subdirs are not supported by uproot
@@ -163,32 +163,6 @@ def read_calib_cache(cachedir, sample, trees, branches, verbose = False) :
 
   return datasets
 
-def read_calib_tuple_2(sample, trees, branches, verbose = False, cachedir = None) : 
-  """ 
-  Read calibration sample from the list of files into numpy array. 
-
-  Args: 
-    sample: tuple in the form (formatstring, numfiles) describing the calibration sample.
-    trees: list of ROOT trees to read from the calibration files 
-    branches: list of branches to read. 
-    cachedir: If not None, store the copy of each input tree into the local cache file 
-              under cachedir (only the selected branches)
-
-  Returns: 
-    2D numpy array, result of concatenation of all calibration samples. 
-    The 1st index of 2D array corresponds to event, 2nd index to the variable from the branches list. 
-  """
-  datasets = []
-  if cachedir : 
-    os.system(f"mkdir -p {cachedir}")
-  filenames = sample
-
-  locations = [ f"{f}:{t}" for f, t in product(filenames, trees) ]
- 
-  print(locations)
-
-  return uproot.concatenate(locations, filter_name = branches, allow_missing = False, library = "pd")[branches].to_numpy()
-
 def calib_transform(x, config, variable) :
   """
   Apply variable transformation to the calibration array. 
@@ -204,13 +178,23 @@ def calib_transform(x, config, variable) :
   for c in config["transform"] : 
     transform_sample += [ eval("lambda x : (" + c + ")") ]
 
-  arr = [ 
-    transform_forward(x[:,0]),    # PID variable
-    transform_sample[0](x[:,1]),  # pT
-    transform_sample[1](x[:,2]),  # eta
-    transform_sample[2](x[:,3] + onp.random.uniform(size = x.shape[0]) - 0.5 ),   # Ntracks (add uniform random +-0.5)
-    x[:,4]  # sWeight
-  ]
+  arr = [ transform_forward(x[:,0]) ]    # PID variable
+  for i,(s,t) in enumerate(zip(config["smear"], transform_sample)) : 
+    # If "smear" not None, do uniform random smearing before transformation
+    if s : 
+      xs = x[:,i+1] + s[0] + (s[1]-s[0])*onp.random.uniform(size = x.shape[0])
+    else : 
+      xs = x[:,i+1]
+    arr += [ t(xs) ]
+  arr += [ x[:,-1] ] # sWeight
+
+#  arr = [ 
+#
+#    transform_sample[0](x[:,1]),  # pT
+#    transform_sample[1](x[:,2]),  # eta
+#    transform_sample[2](x[:,3] + onp.random.uniform(size = x.shape[0]) - 0.5 ),   # Ntracks (add uniform random +-0.5)
+#    x[:,4]  # sWeight
+#  ]
   return np.stack(arr, axis = 1)
 
 def data_transform(x, config) :
@@ -224,11 +208,14 @@ def data_transform(x, config) :
   transform_sample = [ ]
   for c in config["transform"] : 
     transform_sample += [ eval("lambda x : (" + c + ")") ]
-  arr = [ 
-    transform_sample[0](x[:,0]),  # pT
-    transform_sample[1](x[:,1]),  # eta
-    transform_sample[2](x[:,2])   # Ntracks
-  ]
+
+  arr = [ t(x[:,i]) for i, t in enumerate(transform_sample) ]
+
+#  arr = [ 
+#    transform_sample[0](x[:,0]),  # pT
+#    transform_sample[1](x[:,1]),  # eta
+#    transform_sample[2](x[:,2])   # Ntracks
+#  ]
   return np.stack(arr, axis = 1)
 
 def store_to_cache(variable, config, max_files = 0, verbose = False, cachedir = None) : 
@@ -244,20 +231,28 @@ def store_to_cache(variable, config, max_files = 0, verbose = False, cachedir =
   sample = config['sample']
   trees = config['trees']
   calib_branches = [variable["branch"]] + config["branches"]
-  if not cachedir : 
-    calib_cache_dirname = config["calib_cache_dirname"]
-  else : 
-    calib_cache_dirname = cachedir
+  calib_cache_dirname = cachedir
+
+  var_cut = variable.get("cut", None)
+  sample_cut = config.get("cut", None)
+  cut = var_cut
+  if sample_cut:
+    if cut : cut = f"({cut}) & ({sample_cut})"
+    else : cut = sample_cut
+
+  if verbose: 
+    print(f"Cut: {cut}")
 
   if max_files == 0 : 
-    raw_data = read_calib_tuple(sample, trees, calib_branches, verbose = verbose, cachedir = calib_cache_dirname)
+    raw_data = read_calib_tuple(sample, trees, calib_branches, verbose = verbose, cachedir = calib_cache_dirname, cut = cut)
   else : 
-    raw_data = read_calib_tuple(sample[:max_files], trees, calib_branches, verbose = verbose, cachedir = calib_cache_dirname)
+    raw_data = read_calib_tuple(sample[:max_files], trees, calib_branches, verbose = verbose, cachedir = calib_cache_dirname, cut = cut)
   print(f"Read {len(raw_data)} calibration subsamples from remote storage.")
 
 def get_or_create_template(sample_name, dataset_name, variable_name, 
                     variable, config, kernels = None, use_calib_cache = False, control_plots = False, 
-                    interactive_plots = False, prefix = ".", max_files = 0, verbose = False, cachedir = None) : 
+                    interactive_plots = False, local_prefix = ".", global_prefix = None, 
+                    max_files = 0, verbose = False, cachedir = None) : 
   """
 
   """
@@ -268,42 +263,60 @@ def get_or_create_template(sample_name, dataset_name, variable_name,
     "max_files" : max_files, 
   }
 
-  keys_to_ignore = ["labels"]
+  # Remove keys that raw (unsmeared) template does not depend on
+  config_keys_to_ignore = ["labels", "template_sigma"]
+  variable_keys_to_ignore = ["template_sigma"]
 
-  for k in keys_to_ignore : hash_config["config"].pop(k)
+  for k in config_keys_to_ignore : hash_config["config"].pop(k)
+  for k in variable_keys_to_ignore : hash_config["variable"].pop(k)
 
   config_hash = sha1(str(hash_config).encode('utf-8')).hexdigest()
 
   print(f"Template configuration hash : {config_hash}")
 
-  template_storage = f"{prefix}/{sample_name}/{dataset_name}/{variable_name}/{config_hash}/"
-  calib_cache_dirname = f"{cachedir}/{sample_name}/{dataset_name}/{variable_name}/"
+  if cachedir is None : 
+    calib_cache_dirname = None 
+  else : 
+    calib_cache_dirname = f"{cachedir}/{sample_name}/{dataset_name}/{variable_name}/"
 
+  template_storage = f"{global_prefix}/{sample_name}/{dataset_name}/{variable_name}/{config_hash}/"
   try: 
 
-    f = open(template_storage + "/config.json")
-    f.close()
-    data = onp.load(template_storage + "/template.npz", allow_pickle = True)
-    template = data["arr_0"], data["arr_1"], data["arr_2"]
+    #f = open(template_storage + "/config.json")
+    #f.close()
+    #data = onp.load(template_storage + "/template.npz", allow_pickle = True)
+    #template = data["arr_0"], data["arr_1"], data["arr_2"], data["arr_3"]
+    template = load_template(template_storage + "/template.root")
 
     print(f"Read template from {template_storage}")
 
   except FileNotFoundError : 
 
-    print(f"Template in {template_storage} not found, will create one. Be patient... ")
+    print(f"Template in global storage {template_storage} not found, will try local storage.")
 
-    os.system(f"mkdir -p {template_storage}")
-    if not use_calib_cache : 
-      os.system(f"mkdir -p {calib_cache_dirname}")
+    template_storage = f"{local_prefix}/{sample_name}/{dataset_name}/{variable_name}/{config_hash}/"
+    try : 
 
-    template = create_template(variable, config, kernels = kernels, use_calib_cache = use_calib_cache, 
+      template = load_template(template_storage + "/template.root")
+      print(f"Read template from {template_storage}")
+
+    except FileNotFoundError : 
+
+      print(f"Template in local storage {template_storage} not found, will create one. Be patient... ")
+
+      os.system(f"mkdir -p {template_storage}")
+      if (not use_calib_cache) and calib_cache_dirname : 
+        os.system(f"mkdir -p {calib_cache_dirname}")
+
+      template = create_template(variable, config, kernels = kernels, use_calib_cache = use_calib_cache, 
                     control_plots = control_plots, interactive_plots = interactive_plots, 
                     prefix = template_storage, max_files = max_files, verbose = verbose, cachedir = calib_cache_dirname)
 
-    f = open(template_storage + "/config.json", "w")
-    json.dump(hash_config, f, indent = 4)
-    f.close()
-    onp.savez(template_storage + "/template.npz", template[0], template[1], template[2] )
+      f = open(template_storage + "/config.json", "w")
+      json.dump(hash_config, f, indent = 4)
+      f.close()
+      #onp.savez_compressed(template_storage + "/template.npz", template[0], template[1], template[2], template[3] )
+      save_template(template_storage + "/template.root", template)
 
   return template
 
@@ -321,7 +334,7 @@ def create_template(variable, config, kernels = None, use_calib_cache = False, c
     interactive_plots: if True, open control plots in interactive mode, if False, only store them to files. 
     prefix: prefix for control plots (e.g. --prefix="subdir/"). 
     max_files: Maximum number of calibration files to load (0 for unlimited)
-    
+
   Returns: 
     template structure to be used for resampling. 
   """
@@ -337,7 +350,7 @@ def create_template(variable, config, kernels = None, use_calib_cache = False, c
   if kernels : 
     template_sigma = kernels
   else : 
-    template_sigma = [variable["template_sigma"]] + config["template_sigma"]
+    template_sigma = [variable["template_sigma"]["default"]] + config["template_sigma"]["default"]
   max_weights = config["max_weights"]
 
   onp.random.seed(1)  # To make variable transformation of the calibration sample (calib_transform) deterministic
@@ -349,10 +362,18 @@ def create_template(variable, config, kernels = None, use_calib_cache = False, c
       data = read_calib_cache(cachedir, sample[:max_files], trees, calib_branches, verbose = verbose)
     print(f"Read {len(data)} calibration subsamples from local cache.")
   else :
+
+    var_cut = variable.get("cut", None)
+    sample_cut = config.get("cut", None)
+    cut = var_cut
+    if sample_cut:
+      if cut : cut = f"({cut}) & ({sample_cut})"
+      else : cut = sample_cut
+
     if max_files == 0 : 
-      data = read_calib_tuple(sample, trees, calib_branches, verbose = verbose, cachedir = cachedir)
+      data = read_calib_tuple(sample, trees, calib_branches, verbose = verbose, cachedir = cachedir, cut = cut)
     else : 
-      data = read_calib_tuple(sample[:max_files], trees, calib_branches, verbose = verbose, cachedir = cachedir)
+      data = read_calib_tuple(sample[:max_files], trees, calib_branches, verbose = verbose, cachedir = cachedir, cut = cut)
 
     #print(data)
     print(f"Read {len(data)} calibration subsamples from remote storage.")
@@ -375,15 +396,17 @@ def create_template(variable, config, kernels = None, use_calib_cache = False, c
 
   weights = []
   if max_weights : 
+    print(f"Starting to calculate flattening weights")
     histograms = de.create_histograms_vector(data, ranges = ranges, bins = normalise_bins, weights = weights1)[1:]
     for d,w in zip(data, weights1) : 
-      weights2 = de.reweight(d[:,:-1], histograms, max_weights = max_weights, weights = w)
+      weights2 = de.reweight(d[:,1:-1], histograms, max_weights = max_weights, weights = w)
       weights += [ w*weights2 ]
+    if (verbose) : print(f"Weights vector: {weights[0]}")
   else : 
     weights = weights1
 
   print(f"Creating normaliser structure")
-  normaliser = de.create_normaliser_vector(data, ranges = ranges, bins = normalise_bins, weights = weights1)
+  normaliser = de.create_normaliser_vector(data, ranges = ranges, bins = normalise_bins, weights = weights)
 
   print(f"Starting to normalise data array")
   norm_data = []
@@ -394,79 +417,103 @@ def create_template(variable, config, kernels = None, use_calib_cache = False, c
   #unnorm_data = de.unnormalise(norm_data, normaliser, normalise_methods)
 
   counts = None
+  counts2 = None
   edges = None
   for i,(nd,w) in enumerate(zip(norm_data, weights)) : 
     sys.stdout.write(f"\rFilling histogram for subsample {i+1}/{len(norm_data)}")
     sys.stdout.flush()
     c, e = onp.histogramdd(nd, bins = template_bins, range = normalise_ranges, weights = w)
+    c2, e2 = onp.histogramdd(nd, bins = template_bins, range = normalise_ranges, weights = w**2)
     if counts is None : 
       counts = c
+      counts2 = c2
       edges = e
     else : 
       counts += c
+      counts2 += c2
 
   print("")
-  print(f"Applying Gaussian filter")
-  smooth_counts = gaussian_filter(counts, template_sigma)
 
   if control_plots : 
 
+    print(f"Producing control plots")
+
+    print(f"Applying default Gaussian smearing")
+    smooth_counts = gaussian_filter(counts, template_sigma)
+
     labels = config["labels"]
     names = config["names"]
 
     log = True
 
-    print(f"Producing control plots")
     set_lhcb_style(size = 12, usetex = False)
     #fig, axes = plt.subplots(nrows = 7, ncols = 6, figsize = (12, 9) )
 
     for i in range(len(ranges)) : 
-    
+
       if verbose : print(f"Plots for 1D projection {names[i]}")
-     
+
       with plot(f"{names[i]}_transformed", prefix) as (fig, ax) : 
         plot_distr1d(data, i, bins = 50, range = ranges[i], ax = ax, label = "Transformed " + labels[i], weights = weights1, title = "Transformed distribution")
 
-#    for i in range(len(ranges)) : 
       with plot(f"{names[i]}_weighted", prefix) as (fig, ax) : 
         plot_distr1d(data, i, bins = 50, range = ranges[i], ax = ax, label = "Weighted " + labels[i], weights = weights, title = "Weighted distribution")
 
-#    for i in range(len(ranges)) : 
       with plot(f"{names[i]}_normalised", prefix) as (fig, ax) : 
         plot_distr1d(norm_data, i, bins = 50, range = normalise_ranges[i], ax = ax, label = "Normalised " + labels[i], weights = weights, title = "Normalised distribution")
 
-#    for i,j in [ (0,1), (0,2), (1,2), (0,3), (1,3), (2,3) ] : 
+    if len(ranges) == 3 : 
 
-#    bins = template_bins
+      smooth_proj = {
+      (0, 1) : [np.sum(smooth_counts, 2), edges[0], edges[1]],
+      (0, 2) : [np.sum(smooth_counts, 1), edges[0], edges[2]],
+      (1, 2) : [np.sum(smooth_counts, 0), edges[1], edges[2]],
+      }
 
-    smooth_proj = {
+      n1,n2,n3 = [int(n/2) for n in template_bins]  # Make slices through the central region of the distribution
+
+      data_slices = {
+      (0, 1) : [counts[:,:,n3], edges[0], edges[1]], 
+      (0, 2) : [counts[:,n2,:], edges[0], edges[2]], 
+      (1, 2) : [counts[n1,:,:], edges[1], edges[2]], 
+      }
+
+      smooth_slices = {
+      (0, 1) : [smooth_counts[:,:,n3], edges[0], edges[1]], 
+      (0, 2) : [smooth_counts[:,n2,:], edges[0], edges[2]], 
+      (1, 2) : [smooth_counts[n1,:,:], edges[1], edges[2]], 
+      }
+
+    if len(ranges) == 4 : 
+
+      smooth_proj = {
       (0, 1) : [np.sum(smooth_counts, (2,3)), edges[0], edges[1]],
       (0, 2) : [np.sum(smooth_counts, (1,3)), edges[0], edges[2]],
       (1, 2) : [np.sum(smooth_counts, (0,3)), edges[1], edges[2]],
       (0, 3) : [np.sum(smooth_counts, (1,2)), edges[0], edges[3]],
       (1, 3) : [np.sum(smooth_counts, (0,2)), edges[1], edges[3]],
       (2, 3) : [np.sum(smooth_counts, (0,1)), edges[2], edges[3]],
-    }
+      }
 
-    n1,n2,n3,n4 = [int(n/2) for n in template_bins]  # Make slices through the central region of the distribution
+      n1,n2,n3,n4 = [int(n/2) for n in template_bins]  # Make slices through the central region of the distribution
 
-    data_slices = {
+      data_slices = {
       (0, 1) : [counts[:,:,n3,n4], edges[0], edges[1]], 
       (0, 2) : [counts[:,n2,:,n4], edges[0], edges[2]], 
       (1, 2) : [counts[n1,:,:,n4], edges[1], edges[2]], 
       (0, 3) : [counts[:,n2,n3,:], edges[0], edges[3]], 
       (1, 3) : [counts[n1,:,n3,:], edges[1], edges[3]], 
       (2, 3) : [counts[n1,n2,:,:], edges[2], edges[3]], 
-    }
+      }
 
-    smooth_slices = {
+      smooth_slices = {
       (0, 1) : [smooth_counts[:,:,n3,n4], edges[0], edges[1]], 
       (0, 2) : [smooth_counts[:,n2,:,n4], edges[0], edges[2]], 
       (1, 2) : [smooth_counts[n1,:,:,n4], edges[1], edges[2]], 
       (0, 3) : [smooth_counts[:,n2,n3,:], edges[0], edges[3]], 
       (1, 3) : [smooth_counts[n1,:,n3,:], edges[1], edges[3]], 
       (2, 3) : [smooth_counts[n1,n2,:,:], edges[2], edges[3]], 
-    }
+      }
 
     for i,j in smooth_proj.keys() : 
 
@@ -492,7 +539,70 @@ def create_template(variable, config, kernels = None, use_calib_cache = False, c
     #plt.tight_layout(pad=1., w_pad=1., h_pad=0.5)
     if interactive_plots : plt.show()
 
-  return smooth_counts, edges, normaliser
+  return counts.astype(np.float32), np.sqrt(counts2).astype(np.float32), edges, normaliser
+
+def save_template(filename, template) : 
+  """
+    Save PID template into ROOT file 
+    Args: 
+      filename: output ROOT file name
+      template: PID template (in the format as provided by e.g. create_template)
+  """
+
+  counts, counts2, edges, norm = template
+
+  with uproot3.recreate(filename, compression=uproot3.ZLIB(4)) as of :
+
+    of["counts_data"] = uproot3.newtree( { "n" : "float32", "n2" : "float32" } )
+    of["counts_data"].extend({ "n"  : counts.flatten("F"), 
+                               "n2" : counts2.flatten("F") } )
+    of["counts_shape"] = uproot3.newtree( { "i" : "int32"} )
+    of["counts_shape"].extend({ "i" : counts.shape })
+
+#    of["counts2_data"] = uproot3.newtree( { "n" : "float32"} )
+#    of["counts2_data"].extend({ "n" : counts2.flatten("F") })
+#    of["counts2_shape"] = uproot3.newtree( { "i" : "int32"} )
+#    of["counts2_shape"].extend({ "i" : counts2.shape })
+
+    for i,e in enumerate(edges) : 
+      of[f"edges_{i}_data"] = uproot3.newtree( { "n" : "float32"} )
+      of[f"edges_{i}_data"].extend({ "n" : e })
+
+    for i,(cnt, edg) in enumerate(norm) : 
+      of[f"norm_cnt_{i}_data"] = uproot3.newtree( { "n" : "float32"} )
+      of[f"norm_cnt_{i}_data"].extend({ "n" : cnt })
+
+      of[f"norm_edg_{i}_data"] = uproot3.newtree( { "n" : "float32"} )
+      of[f"norm_edg_{i}_data"].extend({ "n" : edg })
+
+def load_template(filename) : 
+  """
+    Load PID template from the ROOT file
+    Args: 
+      filename: Name of the ROOT file or its URL (opened via uproot)
+    
+    Returns: 
+      PID template in the same format as provided by create_template
+  """
+  with uproot.open(filename) as f :
+    counts_data = f["counts_data"]["n"].array(library = "np")
+    counts_shape = f["counts_shape"]["i"].array(library = "np")
+    counts2_data = f["counts_data"]["n2"].array(library = "np")
+
+    counts = counts_data.reshape(counts_shape, order = "F")
+    counts2 = counts2_data.reshape(counts_shape, order = "F")
+
+    edges = []
+    norm = []
+    for i in range(len(counts.shape)) : 
+      edges_data = f[f"edges_{i}_data"]["n"].array(library = "np")
+      edges += [ edges_data ]
+      norm_cnt_data = f[f"norm_cnt_{i}_data"]["n"].array(library = "np")
+      norm_edg_data = f[f"norm_edg_{i}_data"]["n"].array(library = "np")
+      norm += [ (norm_cnt_data, norm_edg_data) ]
+
+    return counts, counts2, edges, norm
+
 
 def resample_data(data, config, variable, template, chunk_size = 50000, verbose = False) : 
   """
