@@ -13,8 +13,9 @@ from ..datasets import legacy_run2_pidcalib_datasets as ds
 from itertools import product
 
 def trees(year) : 
-  return [ ("Lam0" + ("LL" if year in ["2017", "2018"] else "") + \
-            f"_{i}{j}Tuple/DecayTuple") for i, j in product(["", "HPT_", "VHPT_"], ["P", "Pbar"]) ]
+  #return [ ("Lam0" + ("LL" if year in ["2017", "2018"] else "") + \
+  #            f"_{i}{j}Tuple/DecayTuple") for i, j in product(["", "HPT_", "VHPT_"], ["P", "Pbar"]) ]
+  return [ (f"Lam0LL_{i}{j}Tuple/DecayTree") for i, j in product(["", "HPT_", "VHPT_"], ["P", "Pbar"]) ]
 
 sample = {
   f"{pol}_{year}" : { 
@@ -22,17 +23,17 @@ sample = {
     "branches" : [ 
       "probe_Brunel_PT", 
       "probe_Brunel_ETA", 
-      "nTracks_Brunel" if year in ["2017", "2018"] else "nTracks", 
+      "nTracks_Brunel", 
       "probe_sWeight" 
     ], 
     "trees" : trees(year), 
+    "smear" : [ None, None, (-0.5, 0.5) ], 
     "transform" : [ 
       "np.log10(x)",   # pT transformation
       "x",             # eta
       "np.log10(x)",   # number of tracks
     ], 
     "calib_cache_branches" : ["pt", "eta", "ntr", "sw"], 
-    "calib_cache_dirname" : "calib_cache", 
     "data_ranges" : [ (2.4, 4.4), (2.0, 5.0), (0.7, 3.1) ], 
     "labels" : [r"PID", r"$p_T$", r"$\eta$", r"$N_{\rm tr}$"],    # Labels for plots
     "names" : ["pid", "pt", "eta", "ntr"],                        # Short names of each variable for plot file names
@@ -40,8 +41,11 @@ sample = {
     "normalise_bins" : [100, 100, 100], 
     "normalise_methods" : ["flatten", "flatten", "flatten"], 
     "normalise_ranges" : 3*[ (0., 1.) ], 
-    "template_bins" : [100, 100, 50], 
-    "template_sigma" : [5., 5., 5.], 
+    "template_bins" : [70, 70, 20], 
+    "template_sigma" : {
+      "default" : [2., 4., 4.], 
+      "syst1"   : [3., 6., 6.], 
+    }
   } 
   for pol in ["MagUp", "MagDown"] for year in ["2015", "2016", "2017", "2018"]
 }
