@@ -36,6 +36,8 @@ def read_array_filtered(tree, branches, selection = None, sel_branches = []) :
     2D numpy array, where 1st index corresponds to event, 2nd to the variable 
     (in the order given by branches list)
   """
+
+  '''
   arrays = []
   for data in tree.pandas.iterate(branches = branches + sel_branches) : 
       if selection : df = data.query(selection)
@@ -43,6 +45,8 @@ def read_array_filtered(tree, branches, selection = None, sel_branches = []) :
       arr = df[list(branches)].to_numpy()
       arrays += [ arr ]
   return onp.concatenate(arrays, axis = 0)
+  '''
+  return t.arrays(branches, cut = selection, library = "pd")[list(branches)].to_numpy()
 
 def read_array(tree, branches) : 
   """
