@@ -16,7 +16,7 @@ from jax import numpy as np
 import numpy as onp
 import matplotlib
 import matplotlib.pyplot as plt
-import uproot3 as uproot
+import uproot
 import functools
 from scipy.ndimage import gaussian_filter
 
@@ -26,6 +26,65 @@ from .plotting import plot_distr1d_comparison, set_lhcb_style, plot_hist2d, plot
 from .tuples import read_array_filtered, write_array, read_array
 from .resampling import get_samples, get_variables
 
+def validate(
+    input = None, 
+    intree = None, 
+    sample = None, 
+    dataset = None, 
+    variable = None, 
+    weight = None, 
+    branches = None, 
+    pidgen = "pidgen", 
+    piddata = "pid", 
+    output = None, 
+    verbose = False, 
+    interactive = False, 
+    selection = None
+  ) : 
+
+  input_branches = branches.split(":") + [pidgen, piddata, weight]
+
+  pp = pprint.PrettyPrinter(indent = 4)
+
+  config = get_samples()[sample][dataset]
+  vardef = get_variables()[variable]
+
+  if (verbose) : 
+    print(f"Calibration sample config: {pp.pformat(config)}")
+    print(f"Variable definition: {pp.pformat(vardef)}")
+
+  f = uproot.open(input)
+  t = f[intree]
+  branches = t.keys()
+  if (verbose) : print (f"List of all input tree branches: {pp.pformat(branches)}")
+  #input_data = read_array_filtered(t, input_branches, selection = selection, sel_branches = ["pidstat"])
+  input_data = t.arrays(input_branches, selection, library = "pd")[input_branches].to_numpy()
+  if (verbose) : print (f"Input data array shape: {input_data.shape}")
+
+#  transform_forward = lambda x: 1.-(1.-x)**0.2
+  transform_forward = eval("lambda x : (" + vardef["transform_forward"] + ")")
+
+  pidgen_tr = transform_forward(input_data[:,-3])
+  piddata_tr = transform_forward(input_data[:,-2])
+  sw = input_data[:,-1]
+
+  if (verbose) : print(f"Array of sWeights: {sw}")
+
+  label = variable
+
+  set_lhcb_style(size = 12, usetex = False)
+
+  with plot("", output, (5., 4.)) as (fig, ax) : 
+    plot_distr1d_comparison(piddata_tr, pidgen_tr, bins = 100, range = vardef["data_range"], ax = ax, 
+       label = "Transformed PID", 
+       weights = sw, data_weights = sw, 
+       title = "Transformed PID", log = False, pull = True, 
+       legend = ["Original distribution", "Resampled distribution"], 
+       data_alpha = 0.5)
+
+  if (interactive) : plt.show()
+
+
 def main() : 
 
   parser = argparse.ArgumentParser(description = "PIDGen validation script", 
@@ -51,6 +110,8 @@ def main() :
                       help="Original PID branch")
   parser.add_argument('--output', type=str, default = None, 
                       help="Output prefix")
+  parser.add_argument('--selection', type=str, default = None, 
+                      help="Selection string")
   parser.add_argument('--verbose', default = False, action = "store_const", const = True, 
                       help='Enable debug messages')
   parser.add_argument('--interactive', default = False, action = "store_const", const = True, 
@@ -62,54 +123,7 @@ def main() :
     parser.print_help()
     raise SystemExit
 
-  input_tuple = args.input
-  input_tree = args.intree
-  input_branches = args.branches.split(":") + [args.pidgen, args.piddata, args.weight]
-
-  output_tuple = args.output
-
-  sample_name = args.sample
-  dataset_name = args.dataset
-  variable_name = args.variable
-
-  pp = pprint.PrettyPrinter(indent = 4)
-
-  config = get_samples()[sample_name][dataset_name]
-  variable = get_variables()[variable_name]
-
-  if (args.verbose) : 
-    print(f"Calibration sample config: {pp.pformat(config)}")
-    print(f"Variable definition: {pp.pformat(variable)}")
-
-  f = uproot.open(input_tuple)
-  t = f[input_tree]
-  branches = t.keys()
-  if (args.verbose) : print (f"List of all input tree branches: {pp.pformat(branches)}")
-  input_data = read_array_filtered(t, input_branches, selection = "pidstat>1", sel_branches = ["pidstat"])
-  if (args.verbose) : print (f"Input data array shape: {input_data.shape}")
-
-#  transform_forward = lambda x: 1.-(1.-x)**0.2
-  transform_forward = eval("lambda x : (" + variable["transform_forward"] + ")")
-
-  pidgen_tr = transform_forward(input_data[:,-3])
-  piddata_tr = transform_forward(input_data[:,-2])
-  sw = input_data[:,-1]
-
-  if (args.verbose) : print(f"Array of sWeights: {sw}")
-
-  label = variable_name
-
-  set_lhcb_style(size = 12, usetex = False)
-
-  with plot("", args.output, (5., 4.)) as (fig, ax) : 
-    plot_distr1d_comparison(piddata_tr, pidgen_tr, bins = 100, range = variable["data_range"], ax = ax, 
-       label = "Transformed PID", 
-       weights = sw, data_weights = sw, 
-       title = "Transformed PID", log = False, pull = True, 
-       legend = ["Original distribution", "Resampled distribution"], 
-       data_alpha = 0.5)
-
-  if (args.interactive) : plt.show()
+  validate(**vars(args))
 
 if __name__ == "__main__" : 
   main()
